import * as THREE from '../build/three.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { TransformControls } from './jsm/controls/TransformControls.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.shadowMap.enabled = true;
        renderer.shadowMap.type = THREE.PCFShadowMap;
        this.domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        scene.background = new THREE.Color(0xf0f0f0);
        this._scene = scene;

        this._setupCamera();
        this._setupLight();
        this._setupHelper();
        this._setupModel();
        this._setupControls();
        
        window.onresize = this.resize.bind(this);

        this.resize();

        this.render();
    }

    _setupCamera() {
        const aspect = window.innerWidth / window.innerHeight;
        const cameraPersp = new THREE.PerspectiveCamera(50, aspect, 0.01, 30000);
        const cameraOrtho = new THREE.OrthographicCamera(-600*aspect, 600*aspect, 600, -600, 0.01, 30000);
        const currentCamera = cameraPersp;

        currentCamera.position.set(1000, 500, 1000);
        currentCamera.lookAt(0, 200, 0);

        this._cameraPersp = cameraPersp;
        this._cameraOrtho = cameraOrtho;
        this._currentCamera = currentCamera;
    }

    _setupControls() {
        const orbit = new OrbitControls(this._currentCamera, this._renderer.domElement);
        //orbit.update();
        //orbit.addEventListener("change", this.render.bind(this));

        const control = new TransformControls(this._currentCamera, this._renderer.domElement);
        //control.addEventListener("change", this.render.bind(this));
        control.addEventListener("dragging-changed", (event) => {
            this._orbit.enabled = !event.value;
        });

        control.attach(this._mesh);
        this._scene.add(control);

        window.addEventListener("keydown", this._onKeyDown.bind(this));
        window.addEventListener("keyup", this._onKeyUp.bind(this));

        this._orbit = orbit;
        this._control = control;
    }

    _onKeyUp(event) {
        switch(event.keyCode) {
            case 16: // SHIFT
                this._control.setTranslationSnap(null);
                this._control.setRotationSnap(null);
                this._control.setScaleSnap(null);
                break;
        }
    }

    _onKeyDown(event) {
        const control = this._control;

        switch(event.keyCode) {
            case 81: // Q
                control.setSpace(control.space === "local" ? "world" : "local");
                break;
            case 16: // SHIFT
                control.setTranslationSnap(100);
                control.setRotationSnap(THREE.MathUtils.degToRad(15));
                control.setScaleSnap(0.25);
                break;
            case 87: // W
                control.setMode("translate");
                break;
            case 69: // E
                control.setMode("rotate");
                break;
            case 82: // R
                control.setMode("scale");
                break;
            case 67: // C
                const position = this._currentCamera.position.clone();
                this._currentCamera = this._currentCamera.isPerspectiveCamera ? this._cameraOrtho : this._cameraPersp;
                this._currentCamera.position.copy(position);
                this._orbit.object = this._currentCamera;
                this._control.camera = this._currentCamera;
                this._currentCamera.lookAt(this._orbit.target.x, this._orbit.target.y, this._orbit.target.z);
                this.resize();
                break;
            case 86: // V
                const randomFoV = Math.random() + 0.1;
                const randomZoom = Math.random() + 0.1;

                this._cameraPersp.fov = randomFoV * 160;
                this._cameraOrtho.bottom = -randomFoV * 500;
                this._cameraOrtho.top = randomFoV * 500;

                this._cameraPersp.zoom = randomZoom * 5;
                this._cameraOrtho.zoom = randomZoom * 5;

                this.resize();
                break;
            case 187:
            case 107: // +, =, num+
                this._control.setSize(this._control.size + 0.1);
                break;
            case 189:
            case 109: // -, _, num-
                this._control.setSize(Math.max(this._control.size-0.1, 0.1));
                break;
            case 88: // X
                this._control.showX = !this._control.showX;
                break;
            case 89: // Y
                this._control.showY = !this._control.showY;
                break;
            case 90: // Z
                this._control.showZ = !this._control.showZ;
                break;
            case 32: // SPACE
                this._control.enabled = !this._control.enabled;
                break;
        }
    }

    _setupHelper() {
        this._scene.add(new THREE.GridHelper(1000, 10, 0x888888, 0x444444));
    }

    _setupModel() {
        const texture = new THREE.TextureLoader().load("textures/crate.gif", this.render.bind(this));
        texture.anisotropy = this._renderer.capabilities.getMaxAnisotropy();
        
        const geometry = new THREE.BoxGeometry(200, 200, 200);
        const material = new THREE.MeshLambertMaterial( {map: texture, transparent: true} );

        const mesh = new THREE.Mesh(geometry, material);
        this._scene.add(mesh);

        this._mesh = mesh;

    }

    _setupLight() {
        const light = new THREE.DirectionalLight(0xffffff, 2);
        light.position.set(1, 1, 1);
        this._scene.add(light);
    }

    update() {
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._currentCamera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._cameraPersp.aspect = aspect;
        this._cameraPersp.updateProjectionMatrix();
        
        this._cameraOrtho.left = this._cameraOrtho.bottom * aspect;
        this._cameraOrtho.right = this._cameraOrtho.top * aspect;
        this._cameraOrtho.updateProjectionMatrix();

        this._renderer.setPixelRatio(window.devicePixelRatio);
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App());//.render();
}